
//
// Billboard demonstration
// by Jarno van der Linden jvan006@cs.auckland.ac.nz
//
// Demonstrates the use of billboards to draw trees
// The tree billboard is made in a cross configuration
// with two rectangles. One rectangle is on the X plane,
// and the other on the Z plane. Different textures can
// be used to each plane.


#include <cmath>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <ctime>

#include <vector>

#include <gl/glut.h>

#include "controller.h"

#include <iostream>


using namespace std;


#ifndef M_PI
#define M_PI		3.14159265358979323846
#endif




//
// Eight textures are used:
// - TEX_TREE : texture for the tree billboard
// - TEX_SHADOW : texture of the shadow
// - TEX_GROUND : ground texture
// - TEX_SKY_* : sky textures
//
typedef enum { TEX_TREE, TEX_SHADOW, TEX_GROUND,
               TEX_SKY_BACK, TEX_SKY_FRONT, TEX_SKY_LEFT, TEX_SKY_RIGHT, TEX_SKY_UP,
               NUM_TEXTURES } Tex;

GLuint texnames[NUM_TEXTURES] = { 0 };
GLubyte *tex[NUM_TEXTURES] = { NULL };


//
// Controllers
//
class TrivialControlEnvironment : public ControlEnvironment
{
    public:
        float GetHeight(float x, float z)
        {
            return 0;
        };
};
TrivialControlEnvironment controlenv;
ExaminerController examiner(controlenv);
FlyerController flyer(controlenv);


//
// What to draw
//
typedef enum { DRAWMODE_EMPTY=0, DRAWMODE_CARD=0x1, DRAWMODE_CROSS=0x2, DRAWMODEMASK_TYPE=0xf,
               DRAWMODE_WITHTEX=0x10, DRAWMODEMASK_TEX=0xf0,
               DRAWMODE_WITHTRANS=0x100, DRAWMODEMASK_TRANS=0xf00,
               DRAWMODE_WITHPLANEFACING=0x1000, DRAWMODE_WITHFACING=0x2000, DRAWMODEMASK_ORIENT=0xf000,
               DRAWMODE_WITHSHADOW=0x10000, DRAWMODEMASK_SHADOW=0xf0000,
               DRAWMODE_FOREST=0x100000, DRAWMODEMASK_HOWMANY=0xf00000,
} DrawMode;

DrawMode drawmode = DRAWMODE_EMPTY;


// "Names" of the menus
int mainMenu, typeMenu, texMenu, transMenu, orientMenu, shadowMenu, howmanyMenu;


// Tree in a forest
struct Tree
{
    float x, z;
    float widthscale, heightscale;
    float lightness;
    float rot;
};
vector<Tree> forest;
int numtrees = 5000;


inline float max(float a, float b) { return a > b ? a : b; }
inline float min(float a, float b) { return a < b ? a : b; }




//
// Read a PPM file
//
bool ReadPPM(const char* filename, GLubyte*& tex, int& width, int& height)
{
    FILE *fp;
    char str[256];

    fp = fopen(filename, "rb");
    if(!fp)
    {
        fprintf(stderr, "PANIC! Couldn't open texture >%s<\n", filename);
        return false;
    }

    str[0] = '\0';
    fgets(str, 16, fp);
    if(strcmp(str, "P6\n"))
    {
        fclose(fp);
        fprintf(stderr, "PANIC! Texture file not a PPM >%s< : >%s<\n", filename);
        return false;
    }
    do
    {
        fgets(str, 256, fp);
    }
    while(str[0] == '#');

    sscanf(str, "%d %d", &width, &height);
    fgets(str, 256, fp);


    tex = new GLubyte[width*height*3];

    fread(tex, 3, width*height, fp);

    fclose(fp);

    return true;
}


//
// Read in a PPM file and create a texture
//

#define MODE_NONE 0             // No special processing
#define MODE_KEEP 1             // Keep the texture data
#define MODE_BIND 2             // Bind the texture data to a texture
#define MODE_MIPMAP 4           // Create mipmaps for the texture
#define MODE_MAKETRANS 8        // Make texture with alpha transparency
#define MODE_MAKESHADOW 16      // Make a shadow texture

bool ReadTexture(const char* filename, const char* transname, int index, int mode, int wrapmode, int& width, int& height)
{
    unsigned char *tp, *ntp;
    int x, y;
    GLubyte* newtex;
    GLenum srctype;

    if(!ReadPPM(filename, newtex, width, height))
        return false;

    if((mode & (MODE_MAKETRANS | MODE_MAKESHADOW)) && (transname != NULL))
    {
        GLubyte* newtrans;
        int tw, th;
        unsigned char* nttp;

        if(!ReadPPM(transname, newtrans, tw, th))
            return false;
        
        tex[index] = new GLubyte[width*height*4];
        srctype = GL_RGBA;
        tp = tex[index];
        ntp = newtex;
        nttp = newtrans;
        for(y = 0; y < height; y++)
        {
            for(x = 0; x < width; x++)
            {
                if(mode & MODE_MAKESHADOW)
                {
                    *(tp++) = 0;
                    *(tp++) = 0;
                    *(tp++) = 0;
                    *(tp++) = *(nttp++);
                    ntp += 3;
                    nttp += 2;
                }
                else
                {
                    *(tp++) = *(ntp++);
                    *(tp++) = *(ntp++);
                    *(tp++) = *(ntp++);
                    *(tp++) = *(nttp++);
                    nttp += 2;
                }
            }
        }
        delete newtrans;
        delete newtex;
    }
    else if(mode & (MODE_MAKETRANS | MODE_MAKESHADOW))
    {
        tex[index] = new GLubyte[width*height*4];
        srctype = GL_RGBA;
        tp = tex[index];
        ntp = newtex;
        for(y = 0; y < height; y++)
        {
            for(x = 0; x < width; x++)
            {
                if((ntp[0] == 0) && (ntp[1] == 0) && (ntp[2] == 0))
                {
                    if(mode & MODE_MAKESHADOW)
                    {
                        *(tp++) = 0;
                        *(tp++) = 0;
                        *(tp++) = 0;
                        *(tp++) = 0;
                        ntp+=3;
                    }
                    else
                    {
                        *(tp++) = *(ntp++);
                        *(tp++) = *(ntp++);
                        *(tp++) = *(ntp++);
                        *(tp++) = 0;
                    }
                }
                else
                {
                    if(mode & MODE_MAKESHADOW)
                    {
                        *(tp++) = 0;
                        *(tp++) = 0;
                        *(tp++) = 0;
                        *(tp++) = 255;
                        ntp+=3;
                    }
                    else
                    {
                        *(tp++) = *(ntp++);
                        *(tp++) = *(ntp++);
                        *(tp++) = *(ntp++);
                        *(tp++) = 255;
                    }
                }
            }
        }
        delete newtex;
    }
    else
    {
        tex[index] = newtex;
        srctype = GL_RGB;
    }

    if(mode & MODE_BIND)
    {
        glBindTexture(GL_TEXTURE_2D, texnames[index]);
        if(mode & MODE_MIPMAP)
            gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, width, height, srctype, GL_UNSIGNED_BYTE, tex[index]);
        else
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, srctype, GL_UNSIGNED_BYTE, tex[index]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrapmode);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrapmode);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        if(mode & MODE_MIPMAP)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        else
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    }

    if((mode & MODE_KEEP) == 0)
        delete tex[index];

    return true;
}



//
// Make a forest of random trees
//

float frand(const float a, const float b)
{
    return (((float)rand()) / RAND_MAX) * (b-a) + a;
}

void MakeForest(void)
{
    forest.clear();

    srand(42);
    const float range = sqrt(numtrees * 2.f) / 2;
    for(int n = 0; n < numtrees; n++)
    {
        Tree tree;
        tree.x = frand(-range, range);
        tree.z = frand(-range, range);
        tree.widthscale = frand(0.8f, 1.2f);
        tree.heightscale = pow(frand(0.75f, 2.0f), 0.5f);
        tree.lightness = frand(0.5f, 1.25f);
        tree.rot = frand(0, 360);

        forest.push_back(tree);
    }
}


//
// Make a forest of one tree at the origin
//
void MakeTree(void)
{
    forest.clear();

    Tree tree;
    tree.x = tree.z = 0;
    tree.widthscale = tree.heightscale = 1;
    tree.lightness = 1;
    tree.rot = 0;

    forest.push_back(tree);
}



//
// Draw a billboard using a single quad
//
void DrawBillboard(void)
{
    glBegin(GL_QUADS);
        glTexCoord2f(0, 1); glVertex3f(-1, 0, 0);
        glTexCoord2f(0, 0); glVertex3f(-1, 2, 0);
        glTexCoord2f(1, 0); glVertex3f( 1, 2, 0);
        glTexCoord2f(1, 1); glVertex3f( 1, 0, 0);
    glEnd();
}


//
// Draw a quad for a simple shadow on a flat ground in the +Z direction
//
void DrawShadow(void)
{
    glBegin(GL_QUADS);
        glTexCoord2f(0, 1); glVertex3f(-1, 0, 0);
        glTexCoord2f(0, 0); glVertex3f(-1, 0, 1.75);
        glTexCoord2f(1, 0); glVertex3f( 1, 0, 1.75);
        glTexCoord2f(1, 1); glVertex3f( 1, 0, 0);
    glEnd();
}




//
// Draw an empty world
//
void DrawWorld(void)
{
    // Draw using texturing
    glEnable(GL_TEXTURE_2D);

    //
    // The sky box
    //

    // The sky does not influence the depthmap
    glDepthMask(GL_FALSE);
    glPushMatrix();

    // The sky is centred on the viewer
    float pos[3];
    controlenv.GetPos(pos);
    glTranslatef(pos[0], pos[1], pos[2]);
    // Little hack to get the Sun in the sky textures aligned
    // with the shadow direction
    glRotatef(110, 0,1,0);

    glColor3f(0.8,0.8,0.8);
    // Draw the sky as a cube without a bottom
    glBindTexture(GL_TEXTURE_2D, texnames[TEX_SKY_UP]);
    glBegin(GL_QUADS);
        glTexCoord2f(0, 0); glVertex3f(-1,  1, -1);
        glTexCoord2f(1, 0); glVertex3f( 1,  1, -1);
        glTexCoord2f(1, 1); glVertex3f( 1,  1,  1);
        glTexCoord2f(0, 1); glVertex3f(-1,  1,  1);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, texnames[TEX_SKY_FRONT]);
    glBegin(GL_QUADS);
        glTexCoord2f(0, 0); glVertex3f(-1,  1,  1);
        glTexCoord2f(1, 0); glVertex3f( 1,  1,  1);
        glTexCoord2f(1, 1); glVertex3f( 1, -1,  1);
        glTexCoord2f(0, 1); glVertex3f(-1, -1,  1);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, texnames[TEX_SKY_RIGHT]);
    glBegin(GL_QUADS);
        glTexCoord2f(0, 0); glVertex3f( 1,  1,  1);
        glTexCoord2f(1, 0); glVertex3f( 1,  1, -1);
        glTexCoord2f(1, 1); glVertex3f( 1, -1, -1);
        glTexCoord2f(0, 1); glVertex3f( 1, -1,  1);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, texnames[TEX_SKY_BACK]);
    glBegin(GL_QUADS);
        glTexCoord2f(0, 0); glVertex3f( 1,  1, -1);
        glTexCoord2f(1, 0); glVertex3f(-1,  1, -1);
        glTexCoord2f(1, 1); glVertex3f(-1, -1, -1);
        glTexCoord2f(0, 1); glVertex3f( 1, -1, -1);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, texnames[TEX_SKY_LEFT]);
    glBegin(GL_QUADS);
        glTexCoord2f(0, 0); glVertex3f(-1,  1, -1);
        glTexCoord2f(1, 0); glVertex3f(-1,  1,  1);
        glTexCoord2f(1, 1); glVertex3f(-1, -1,  1);
        glTexCoord2f(0, 1); glVertex3f(-1, -1, -1);
    glEnd();

    glPopMatrix();
    glDepthMask(GL_TRUE);


    //
    // The ground plane
    //

    // This could be done as a single very large quad,
    // but splitting it up into smaller pieces appears
    // to give a more stable-looking result.
    glColor3f(1,1,1);
    glBindTexture(GL_TEXTURE_2D, texnames[TEX_GROUND]);
    const int step = 10;
    for(int z = -100; z < 100; z += step)
    {
        glBegin(GL_TRIANGLE_STRIP);
        for(int x = -100; x < 100; x += step)
        {
            glTexCoord2f(x, z); glVertex3f(x, 0, z);
            glTexCoord2f(x, z+step); glVertex3f(x, 0, z+step);
        }
        glEnd();
    }

    glDisable(GL_TEXTURE_2D);
}



//
// Draw a forest of trees
//
void DrawForest(void)
{
    // First draw all the shadows on the ground, then draw all the
    // trees. Although shadows and trees could be drawn together,
    // rendering objects with the same texture together saves a lot
    // of texture context switching. Changing the texture to use
    // can be expensive.

    if((drawmode & DRAWMODEMASK_SHADOW) == DRAWMODE_WITHSHADOW)
    {
        if((drawmode & DRAWMODEMASK_TEX) == DRAWMODE_WITHTEX)
        {
            // Use the shadow texture
            glBindTexture(GL_TEXTURE_2D, texnames[TEX_SHADOW]);
            // Draw using texturing
            glEnable(GL_TEXTURE_2D);
        }

        // Shadows are drawn using blending
        if((drawmode & DRAWMODEMASK_TRANS) == DRAWMODE_WITHTRANS)
            glEnable(GL_BLEND);

        // Shadows don't set depth values
        glDepthMask(GL_FALSE);
        // Shadows are drawn first, and will always appear on top
        // of the ground. So depth testing is not needed.
        // Furthermore, turning it off prevents Z-fighting between
        // the shadow polygons and the ground polygon.
        glDisable(GL_DEPTH_TEST);

        // Set the shadow with transparency set to half, so that any
        // areas in shadow will be darkened by the shadow but not
        // completely black.
        glColor4f(0,0,0,0.5);

        // Draw the shadow for each tree
        for(vector<Tree>::const_iterator ti = forest.begin(); ti != forest.end(); ti++)
        {
            glPushMatrix();
            // Place the shadow at the same position as the tree
            glTranslatef(ti->x, 0, ti->z);
            // To keep things simple the shadow will always stretch
            // out in the +Z direction away from the tree. So scaling
            // the height of the tree causes the shadow to scale in
            // the Z direction.
            glScalef(ti->widthscale, 1, ti->heightscale);

            DrawShadow();

            glPopMatrix();
        }

        // Restore the state to what the rest of the program expects
        glEnable(GL_DEPTH_TEST);
        glDepthMask(GL_TRUE);
        glDisable(GL_BLEND);
        glDisable(GL_TEXTURE_2D);
    }


    if((drawmode & DRAWMODEMASK_TYPE) != DRAWMODE_EMPTY)
    {

        if((drawmode & DRAWMODEMASK_TEX) == DRAWMODE_WITHTEX)
        {
            // Use tree texture
            glBindTexture(GL_TEXTURE_2D, texnames[TEX_TREE]);
            // Draw using texturing
            glEnable(GL_TEXTURE_2D);
        }

        // Trees are drawn using masking
        if((drawmode & DRAWMODEMASK_TRANS) == DRAWMODE_WITHTRANS)
            glEnable(GL_ALPHA_TEST);

        // Draw each tree
        for(vector<Tree>::const_iterator ti = forest.begin(); ti != forest.end(); ti++)
        {
            glPushMatrix();
            // Place the tree
            glTranslatef(ti->x, 0, ti->z);
            // Scale the tree
            glScalef(ti->widthscale, ti->heightscale, ti->widthscale);
            // With the GL_MODULATE texture mode the texture colours
            // are multiplied by the glColor values. This is an easy
            // way to make a texture lighter or darker, or even give
            // it a tint.
            glColor3f(ti->lightness, ti->lightness, ti->lightness);

            // Orient the tree
            float rot = ti->rot;
            if((drawmode & DRAWMODEMASK_ORIENT) == DRAWMODE_WITHPLANEFACING)
            {
                // To make a billboard always face the camera plane,
                // just rotate it the same as the camera. Here, the
                // tree billboards can only change their heading.
                float camrot[3];
                controlenv.GetRot(camrot);
                rot = camrot[1] / M_PI * 180;
            }
            else if((drawmode & DRAWMODEMASK_ORIENT) == DRAWMODE_WITHFACING)
            {
                // When set to always face the camera position, compute
                // the tree rotation so that it faces the point on the
                // ground where the camera is above.
                float campos[3];
                controlenv.GetPos(campos);
                rot = atan2(campos[0] - ti->x, campos[2] - ti->z) / M_PI * 180;
            }
            glRotatef(rot, 0,1,0);

            switch(drawmode & DRAWMODEMASK_TYPE)
            {
                case DRAWMODE_CARD:
                    // Draw one billboard
                    DrawBillboard();
                    break;
                case DRAWMODE_CROSS:
                    // Draw two billboards, one 90 degrees from the other
                    DrawBillboard();
                    glRotatef(90, 0,1,0);
                    DrawBillboard();
                    break;
            }

            glPopMatrix();
        }

        // Restore the state to what the rest of the program expects
        glDisable(GL_ALPHA_TEST);
        glDisable(GL_TEXTURE_2D);
    }
}


//
// GLUT display routine
//
void Display(void)
{
    static clock_t lasttime(clock());
    clock_t thistime = clock();

    // Initialise for new drawing
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    controlenv.Move((thistime - lasttime) / (float)CLOCKS_PER_SEC);
    lasttime = thistime;

    // Remember that in OpenGL you don't move the camera,
    // but you move the rest of the world. Transform to
    // camera-relative coordinates by undoing the position
    // and rotation of the camera in world coordinates.
    float rot[3];
    controlenv.GetRot(rot);
    glRotatef(-rot[2] / M_PI * 180, 0, 0, 1);
    glRotatef(-rot[0] / M_PI * 180, 1, 0, 0);
    glRotatef(-rot[1] / M_PI * 180, 0, 1, 0);
    float pos[3];
    controlenv.GetPos(pos);
    glTranslatef(-pos[0], -pos[1], -pos[2]);

    DrawWorld();

    DrawForest();

    glFlush();
    glutSwapBuffers();
}


void Reshape(int width, int height )
{
    glViewport(0, 0, width, height);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(35, width / (float)height, 0.1, 500);

    glMatrixMode(GL_MODELVIEW);

    controlenv.CentreMouse(width / 2, height / 2);
}


void Idle(void)
{
    glutPostRedisplay();
}


//
// GLUT mouse and keyboard handling
//
void HandleMouseMotion(int x,int y)
{
    controlenv.MouseMoveAbs(x, y);
}

void HandleMouseClick(int button, int state, int x, int y)
{
    if(button == GLUT_LEFT_BUTTON)
        controlenv.MouseButtons((state == GLUT_DOWN), 0, 0);
    else if(button == GLUT_RIGHT_BUTTON)
        controlenv.MouseButtons(0, 0, (state == GLUT_DOWN));
}


//
// GLUT menu handling
//
void TypeMenuFunc(int menuItem)
{
    drawmode = (DrawMode)((drawmode & ~DRAWMODEMASK_TYPE) | menuItem);
    glutPostRedisplay();
}
void TexMenuFunc(int menuItem)
{
    drawmode = (DrawMode)((drawmode & ~DRAWMODEMASK_TEX) | menuItem);
    glutPostRedisplay();
}
void TransMenuFunc(int menuItem)
{
    drawmode = (DrawMode)((drawmode & ~DRAWMODEMASK_TRANS) | menuItem);
    glutPostRedisplay();
}
void ShadowMenuFunc(int menuItem)
{
    drawmode = (DrawMode)((drawmode & ~DRAWMODEMASK_SHADOW) | menuItem);
    glutPostRedisplay();
}
void HowmanyMenuFunc(int menuItem)
{
    drawmode = (DrawMode)((drawmode & ~DRAWMODEMASK_HOWMANY) | menuItem);
    if(menuItem == DRAWMODE_FOREST)
    {
        MakeForest();
        controlenv.SetController(flyer);
    }
    else
    {
        MakeTree();
        controlenv.SetController(examiner);
    }

    glutPostRedisplay();
}
void OrientMenuFunc(int menuItem)
{
    drawmode = (DrawMode)((drawmode & ~DRAWMODEMASK_ORIENT) | menuItem);
    glutPostRedisplay();
}

void CreateMenus(void)
{
    mainMenu = glutCreateMenu(NULL);

    typeMenu = glutCreateMenu(TypeMenuFunc);
    glutAddMenuEntry("Empty", DRAWMODE_EMPTY);
    glutAddMenuEntry("Single card", DRAWMODE_CARD);
    glutAddMenuEntry("Cross", DRAWMODE_CROSS);

    texMenu = glutCreateMenu(TexMenuFunc);
    glutAddMenuEntry("On", DRAWMODE_WITHTEX);
    glutAddMenuEntry("Off", 0);

    transMenu = glutCreateMenu(TransMenuFunc);
    glutAddMenuEntry("On", DRAWMODE_WITHTRANS);
    glutAddMenuEntry("Off", 0);

    orientMenu = glutCreateMenu(OrientMenuFunc);
    glutAddMenuEntry("Camera plane", DRAWMODE_WITHPLANEFACING);
    glutAddMenuEntry("Camera position", DRAWMODE_WITHFACING);
    glutAddMenuEntry("Off", 0);

    shadowMenu = glutCreateMenu(ShadowMenuFunc);
    glutAddMenuEntry("On", DRAWMODE_WITHSHADOW);
    glutAddMenuEntry("Off", 0);

    howmanyMenu = glutCreateMenu(HowmanyMenuFunc);
    glutAddMenuEntry("Single", 0);
    glutAddMenuEntry("Forest", DRAWMODE_FOREST);

    glutSetMenu(mainMenu);
    glutAddSubMenu("Type", typeMenu);
    glutAddSubMenu("Texture", texMenu);
    glutAddSubMenu("Transparency", transMenu);
    glutAddSubMenu("Face camera", orientMenu);
    glutAddSubMenu("Shadow", shadowMenu);
    glutAddSubMenu("How many", howmanyMenu);

    glutAttachMenu(GLUT_MIDDLE_BUTTON);
}


void HandleKeyboardEvent(unsigned char key, int x, int y)
{
    switch(key)
    {
        case 27:
            exit(0);
            break;
        case ']':
            numtrees += 1000;
            MakeForest();
            break;
        case '[':
            numtrees -= 1000;
            numtrees = max(numtrees, 0);
            MakeForest();
            break;
        case 'e': case 'E':
            TypeMenuFunc(0);
            break;
        case 'c':
            TypeMenuFunc(DRAWMODE_CARD);
            break;
        case 'C':
            TypeMenuFunc(DRAWMODE_CROSS);
            break;
        case 'T':
            TexMenuFunc(0);
            break;
        case 't':
            TexMenuFunc(DRAWMODE_WITHTEX);
            break;
        case 'A':
            TransMenuFunc(0);
            break;
        case 'a':
            TransMenuFunc(DRAWMODE_WITHTRANS);
            break;
        case 'p': case 'P':
            OrientMenuFunc(0);
            break;
        case 'f':
            OrientMenuFunc(DRAWMODE_WITHPLANEFACING);
            break;
        case 'F':
            OrientMenuFunc(DRAWMODE_WITHFACING);
            break;
        case 'S':
            ShadowMenuFunc(0);
            break;
        case 's':
            ShadowMenuFunc(DRAWMODE_WITHSHADOW);
            break;
        case '1':
            HowmanyMenuFunc(0);
            break;
        case '2':
            HowmanyMenuFunc(DRAWMODE_FOREST);
            break;
        default:
            break;
    }
    glutPostRedisplay();
}



//
// Main
//
int main(int argc, char* argv[])
{
    //
    // Set up an OpenGL context using GLUT
    //

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);

    glutInitWindowSize(800,600);
    glutInitWindowPosition(50,50);
    glutCreateWindow("Billboard Trees");

    //
    // Set up mouse, keyboard, and menu handling
    //

    glutSetCursor(GLUT_CURSOR_CROSSHAIR);
    glutMouseFunc(HandleMouseClick);
    glutMotionFunc(HandleMouseMotion);
    glutPassiveMotionFunc(HandleMouseMotion);
    glutKeyboardFunc(HandleKeyboardEvent);
    CreateMenus();

    //
    // Set up rendering
    //

    glutReshapeFunc(Reshape);
    glutDisplayFunc(Display);
    glutIdleFunc(Idle);

    //
    // Initialise some OpenGL state
    //

    glClearColor(124/255.0,149/255.0,248/255.0,1);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glAlphaFunc(GL_GREATER, 0.5f);

    glFogi(GL_FOG_MODE,GL_EXP);
    glFogf(GL_FOG_DENSITY, 0.01f);
    GLfloat fog_col[4] = { 0.75f, 0.79f, 0.82f, 0 };
    glFogfv(GL_FOG_COLOR,fog_col);
    glEnable(GL_FOG);

    //
    // Create textures
    //

    glPixelStorei(GL_PACK_ALIGNMENT,1);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    glGenTextures(NUM_TEXTURES, texnames);

    int width, height;
    if(!ReadTexture("images/Tree2.ppm", "images/Tree2_mask.ppm", TEX_TREE, MODE_BIND | MODE_MAKETRANS | MODE_MIPMAP, GL_CLAMP, width, height))
        return EXIT_FAILURE;
    if(!ReadTexture("images/Tree2.ppm", "images/Tree2_mask.ppm", TEX_SHADOW, MODE_BIND | MODE_MAKESHADOW | MODE_MIPMAP, GL_CLAMP, width, height))
        return EXIT_FAILURE;
    if(!ReadTexture("images/ground.ppm", NULL, TEX_GROUND, MODE_BIND | MODE_MIPMAP, GL_REPEAT, width, height))
        return EXIT_FAILURE;
    if(!ReadTexture("images/sky_back.ppm", NULL, TEX_SKY_BACK, MODE_BIND | MODE_MIPMAP, GL_CLAMP, width, height))
        return EXIT_FAILURE;
    if(!ReadTexture("images/sky_front.ppm", NULL, TEX_SKY_FRONT, MODE_BIND | MODE_MIPMAP, GL_CLAMP, width, height))
        return EXIT_FAILURE;
    if(!ReadTexture("images/sky_left.ppm", NULL, TEX_SKY_LEFT, MODE_BIND | MODE_MIPMAP, GL_CLAMP, width, height))
        return EXIT_FAILURE;
    if(!ReadTexture("images/sky_right.ppm", NULL, TEX_SKY_RIGHT, MODE_BIND | MODE_MIPMAP, GL_CLAMP, width, height))
        return EXIT_FAILURE;
    if(!ReadTexture("images/sky_up.ppm", NULL, TEX_SKY_UP, MODE_BIND | MODE_MIPMAP, GL_CLAMP, width, height))
        return EXIT_FAILURE;

    //
    // Set up controllers
    //

    float lookat[3] = { 0, 1, 0 };
    examiner.SetLookAt(lookat);
    float pos[3] = { 0, 1, 4 };
    examiner.SetPos(pos);
    examiner.SetParam(ExaminerController::dist_sensitive_Param, 2);
    flyer.SetParam(FlyerController::accel_sensitive_Param, 4);

    //
    // Start off with one tree and the examiner controller
    //

    MakeTree();
    controlenv.SetController(examiner);

    glutMainLoop();

    return(0);
}
