//
// Controller
// An attempt at a unified mouse control system
// jvan006@cs.auckland.ac.nz
//

#include "controller.h"

#include <math.h>
#include <float.h>


#ifndef M_PI
#define M_PI   3.14159265358979323846f
#endif
#ifndef PI
#define PI M_PI
#endif


ControlEnvironment::ControlEnvironment()
{
    pos[0] = pos[1] = pos[2] = 0;
    rot[0] = rot[1] = rot[2] = 0;
    speed = 0;
    mousex = mousey = 0;
    mousecx = mousecy = 0;
    curcontrol = 0;
}


ControlEnvironment::~ControlEnvironment()
{
    ;
}


void ControlEnvironment::SetPos(float pos_in[3])
{
    if(curcontrol)
        curcontrol->SetPos(pos_in);
    else
        SetControlPos(pos_in);
}


void ControlEnvironment::SetRot(float rot_in[3])
{
    if(curcontrol)
        curcontrol->SetRot(rot_in);
    else
        SetControlRot(rot_in);
}


void ControlEnvironment::SetSpeed(float speed_in)
{
    if(curcontrol)
        curcontrol->SetSpeed(speed_in);
    else
        SetControlSpeed(speed_in);
}


void ControlEnvironment::SetControlPos(float pos_in[3])
{
    pos[0] = pos_in[0];
    pos[1] = pos_in[1];
    pos[2] = pos_in[2];
}


void ControlEnvironment::SetControlRot(float rot_in[3])
{
    int i;

    for(i=0; i<3; i++)
    {
        rot[i] = rot_in[i];
        // Keep rotations between [0..2*M_PI)
        rot[i] = (float)fmod(rot[i], 2*M_PI);
        if(rot[i] < 0)
            rot[i] += 2*M_PI;
    }
}


void ControlEnvironment::SetControlSpeed(float speed_in)
{
    speed = speed_in;
}


void ControlEnvironment::GetPos(float pos_out[3])
{
    pos_out[0] = pos[0];
    pos_out[1] = pos[1];
    pos_out[2] = pos[2];
}


void ControlEnvironment::GetRot(float rot_out[3])
{
    rot_out[0] = rot[0];
    rot_out[1] = rot[1];
    rot_out[2] = rot[2];
}


float ControlEnvironment::GetSpeed()
{
    return speed;
}


void ControlEnvironment::SetParam(int param, float val)
{
    ;
}


float ControlEnvironment::GetParam(int param)
{
    return 0;
}


void ControlEnvironment::SetController(Controller& control)
{
    float oldpos[3], oldrot[3], oldspeed;

    if(&(control.GetControlEnvironment()) == this)
    {
        curcontrol = &control;
        GetPos(oldpos);
        GetRot(oldrot);
        oldspeed = GetSpeed();

        SetPos(oldpos);
        SetRot(oldrot);
        SetSpeed(oldspeed);

        Move(0);
    }
}


Controller& ControlEnvironment::GetController()
{
    return *curcontrol;
}


const char *ControlEnvironment::GetControllerName()
{
    if(curcontrol)
        return curcontrol->GetName();
    else
        return "(none)";
}


void ControlEnvironment::MouseMoveAbs(int x, int y)
{
    MouseMove((x - mousecx) - mousex, (y - mousecy) - mousey);
}

void ControlEnvironment::CentreMouse(int x, int y)
{
    MouseMove(mousecx - x, mousecy - y);
    mousecx = x;
    mousecy = y;
}

void ControlEnvironment::MouseMove(int deltax, int deltay)
{
    mousex += deltax;
    mousey += deltay;
    if(curcontrol)
        curcontrol->MouseMove(deltax, deltay);
}

void ControlEnvironment::GetMousePos(int& x, int& y)
{
    x = mousex;
    y = mousey;
}


void ControlEnvironment::MouseButtons(int leftbutton, int middlebutton, int rightbutton)
{
    if(curcontrol)
        curcontrol->MouseButtons(leftbutton, middlebutton, rightbutton);
}


void ControlEnvironment::MouseWheel(int delta)
{
    if(curcontrol)
        curcontrol->MouseWheel(delta);
}


void ControlEnvironment::Move(float deltatime)
{
    if(curcontrol)
        curcontrol->Move(deltatime);
}



Controller::Controller(ControlEnvironment& controlenv)
    : controlenv(&controlenv)
{
    ;
}

Controller::~Controller()
{
    ;
}


void Controller::SetPos(float pos_in[3])
{
    if(controlenv)
        controlenv->SetControlPos(pos_in);
}


void Controller::SetRot(float rot_in[3])
{
    if(controlenv)
        controlenv->SetControlRot(rot_in);
}


void Controller::SetSpeed(float speed_in)
{
    if(controlenv)
        controlenv->SetControlSpeed(speed_in);
}


void Controller::GetPos(float pos_out[3])
{
    if(controlenv)
        controlenv->GetPos(pos_out);
    else
        pos_out[0] = pos_out[1] = pos_out[2] = 0;
}


void Controller::GetRot(float rot_out[3])
{
    if(controlenv)
        controlenv->GetRot(rot_out);
    else
        rot_out[0] = rot_out[1] = rot_out[2] = 0;
}


float Controller::GetSpeed()
{
    if(controlenv)
        return controlenv->GetSpeed();
    else
        return 0;
}


void Controller::SetParam(int param, float val)
{
    ;
}


float Controller::GetParam(int param)
{
    return 0;
}


ControlEnvironment& Controller::GetControlEnvironment()
{
    return *controlenv;
}


float Controller::GetHeight(float x, float z)
{
    if(controlenv)
        return controlenv->GetHeight(x, z);
    else
        return 0;
}


StriderController::StriderController(ControlEnvironment& controlenv)
    : Controller(controlenv)
{
    // Initial state
    reqheight = 0;
    reqspeed = 0;
    inc = 0;
    lasty = 0;
    twist = 0;
    twisting = 0;

    // Initial parameters
    deadzone = 100.f;
    speedsensitive = 1.f;
    forwardsetspeedmax = 30.f;
    backwardsetspeedmax = 30.f;
    headingsensitive = 0.3f;
    heightsensitive = 10.f;
    looksensitive = 6.25f;
    forwardspeedmax = 100.f;
    backwardspeedmax = 100.f;
    setheightmax = 50.f;
    eyeheightmin = 0.2f;
    twistrestore = 0.05f;
}


StriderController::~StriderController()
{
    ;
}


void StriderController::MouseMove(int deltax, int deltay)
{
    float rot[3];
    float newspeed;

    Controller::GetRot(rot);
    rot[1] -= twist;

    int mx = 0, my = 0;
    GetControlEnvironment().GetMousePos(mx, my);

    // When speed is zero, only begin moving if mouse
    // has been moved at least beyond a threshold distance
    // from the place where the mouse was when speed got
    // to zero.
    if((reqspeed == 0) && (abs(my - lasty) < deadzone))
        ;
    else
    {
        // Compute new speed based on how much the mouse moved
        newspeed = reqspeed - deltay * speedsensitive / 100.f;

        if(((reqspeed > 0) && (newspeed <= 0)) ||
           ((reqspeed < 0) && (newspeed >= 0)))
        {
            // Reached or cross zero speed
            // Make speed exactly zero
            reqspeed = 0;
            lasty = my;
        }
        else
        {
            reqspeed = newspeed;
        }

        // Limit the maximum speed
        if(reqspeed > forwardsetspeedmax)
            reqspeed = forwardsetspeedmax;
        else if(reqspeed < -backwardsetspeedmax)
            reqspeed = -backwardsetspeedmax;
    }

    // If twisting, change twist with mx
    // else change heading

    if(twisting)
    {
        twist -= deltax * headingsensitive / 100.f;
    }
    else
    {
        // Compute new heading (rot around Y) based on how much
        // the mouse moved.
        rot[1] -= deltax * headingsensitive / 100.f;
    }

    rot[1] += twist;
    Controller::SetRot(rot);
}


void StriderController::MouseButtons(int leftbutton, int middlebutton, int rightbutton)
{
    // Left button causes viewer to move up, right button down
    inc = (float)(leftbutton - rightbutton) * heightsensitive;
    twisting = middlebutton;
}


void StriderController::MouseWheel(int delta)
{
    float rot[3];

    Controller::GetRot(rot);

    // Wheel controls look up/down
    rot[0] += delta * looksensitive / 100.f;

    Controller::SetRot(rot);
}


void StriderController::Move(float deltatime)
{
    float pos[3], rot[3], speed;
    float scale, height;
    float aim[3];

    Controller::GetPos(pos);
    Controller::GetRot(rot);

    // rot[0] = pitch, rot[1] = heading, rot[2] = bank (not used)

    // remove twist
    rot[1] -= twist;

    // Compute normalised (???) aim vector
    aim[0] = -sinf(rot[1]);
    aim[1] = inc;
    aim[2] = -cosf(rot[1]);

    // How high above the terrain are we now?
    height = Controller::GetHeight(pos[0], pos[2]);

    // Computing the movement scale is mostly fudge made to feel right
    // It is relative to height above terrain
    scale = (pos[1] - height + 1) / 2.f;
    if(scale < 1) scale = 1;
    // There may be some psychological reason why using sqrt works
    scale = sqrtf(scale);

    // Scale the speed, within limits
    speed = reqspeed * scale;
    if(speed > forwardspeedmax)
        speed = forwardspeedmax;
    else if(speed < -backwardspeedmax)
        speed = -backwardspeedmax;

    // Change position
    pos[0] += aim[0] * speed * deltatime;
    pos[2] += aim[2] * speed * deltatime;
    // How fast to move up/down is not dependent on the speed, only scale
    reqheight += aim[1] * scale * deltatime;

    // Avoid collisions with extraterrestrial bodies
    if(reqheight > setheightmax)
        reqheight = setheightmax;

    // The actual position is the desired position possibly modified
    // temporarily for terrain avoidance
    pos[1] = reqheight;

    // Get terrain height at current position, plus some delta to avoid
    // scraping the terrain
    height = Controller::GetHeight(pos[0], pos[2]);
    height += eyeheightmin;
    if(pos[1] < height)
    {
        pos[1] = height;
        // If the user is changing the height manually while terrain
        // avoidance is in effect, set the new desired height to the
        // current actual height. This avoids the apparent lack of
        // response which would otherwise occur as the desired height
        // catches up with the actual height
        if(inc)
            reqheight = height;
    }

    // Strider has no banking
    rot[2] = 0;

    // Add in head twist
    if((!twisting) && (twist != 0))
    {
        float dtwist = (float)fabs(twist) * twistrestore;
        if(dtwist < 0.1f)
            dtwist = 0.1f;
        if(fabs(twist) < dtwist)
            twist = 0;
        else if(twist < 0)
            twist += dtwist;
        else if(twist > 0)
            twist -= dtwist;
    }

    rot[1] += twist;

    Controller::SetPos(pos);
    Controller::SetRot(rot);
    Controller::SetSpeed(speed);
}


void StriderController::SetPos(float pos_in[3])
{
    float height;

    // When manually setting a new position, make the given y position
    // also the requested height
    // Hack: only change reqheight is new height is significantly above terrain
    height = Controller::GetHeight(pos_in[0], pos_in[2]);
    height += eyeheightmin;
    if(pos_in[1] > (height + FLT_EPSILON*8))
        reqheight = pos_in[1];
    Controller::SetPos(pos_in);
}


void StriderController::SetRot(float rot_in[3])
{
    float rot[3];

    // When manually setting a new rotation, strip out banking
    rot[0] = rot_in[0];
    rot[1] = rot_in[1];
    rot[2] = 0;
    Controller::SetRot(rot);
}


void StriderController::SetSpeed(float speed)
{
    // When manually setting a new speed, make it reqspeed
    reqspeed = speed;
    // TODO: compute a reqspeed, taking speed to be actual speed
    Controller::SetSpeed(speed);
}


const char *StriderController::GetName()
{
    return "Strider";
}


float *StriderController::GetParamP(int param)
{
    switch(param)
    {
        case deadzone_Param:
            return &deadzone;
        case speed_sensitive_Param:
            return &speedsensitive;
        case forward_set_speed_max_Param:
            return &forwardsetspeedmax;
        case backward_set_speed_max_Param:
            return &backwardsetspeedmax;
        case heading_sensitive_Param:
            return &headingsensitive;
        case height_sensitive_Param:
            return &heightsensitive;
        case look_sensitive_Param:
            return &looksensitive;
        case forward_speed_max_Param:
            return &forwardspeedmax;
        case backward_speed_max_Param:
            return &backwardspeedmax;
        case set_height_max_Param:
            return &setheightmax;
        case eye_height_min_Param:
            return &eyeheightmin;
    }

    return 0;
}


void StriderController::SetParam(int param, float val)
{
    float *paramp;

    paramp = GetParamP(param);
    if(paramp)
        *paramp = val;
    else
        Controller::SetParam(param, val);
}


float StriderController::GetParam(int param)
{
    float *paramp;

    paramp = GetParamP(param);
    if(paramp)
        return *paramp;
    else
        return Controller::GetParam(param);
}




FlyerController::FlyerController(ControlEnvironment& controlenv)
    : Controller(controlenv)
{
    // Initial parameters
    pitchsensitive = 0.3f;
    banksensitive = 0.3f;
    pitchmax = M_PI/2.f;
    bankmax = M_PI/2.f;
    accelsensitive = 50.f;
    forwardspeedmax = 30.f;
    backwardspeedmax = 30.f;
    setheightmax = 50.f;
    eyeheightmin = 0.2f;
}


FlyerController::~FlyerController()
{
    ;
}


void FlyerController::MouseMove(int deltax, int deltay)
{
    float rot[3];

    Controller::GetRot(rot);

    // Change pitch
    // Limit the pitch angle
    int mx, my;
    GetControlEnvironment().GetMousePos(mx, my);
    rot[0] = (float)my * pitchsensitive / 100.f;
    //rot[0] += (float)deltay * pitchsensitive / 100.f;
    rot[0] = (float)fmod(rot[0], 2*M_PI);
    if(rot[0] < 0)
        rot[0] += 2*M_PI;
    if((rot[0] < M_PI) && (rot[0] > pitchmax))
        rot[0] = pitchmax;
    else if((rot[0] > M_PI) && (rot[0] < (2*M_PI-pitchmax)))
        rot[0] = 2*M_PI-pitchmax;

    // Change bank
    rot[2] = -(float)mx *banksensitive / 100;
    //rot[2] -= (float)deltax *banksensitive / 100;
    rot[2] = (float)fmod(rot[2], 2*M_PI);
    if(rot[2] < 0)
        rot[2] += 2*M_PI;
    if((rot[2] < M_PI) && (rot[2] > bankmax))
        rot[2] = bankmax;
    else if((rot[2] > M_PI) && (rot[2] < (2*M_PI-bankmax)))
        rot[2] = 2*M_PI-bankmax;

    Controller::SetRot(rot);
}


void FlyerController::MouseButtons(int leftbutton, int middlebutton, int rightbutton)
{
    // Left button causes viewer to accelerate
    // Right button decelerates
    accel = (float)(leftbutton - rightbutton) * accelsensitive;
}


void FlyerController::MouseWheel(int delta)
{
    // Not used
    ;
}


void FlyerController::Move(float deltatime)
{
    float rot[3];
    float pos[3];
    float speed;
    float height;
    float aim[3];

    Controller::GetRot(rot);
    Controller::GetPos(pos);
    speed = Controller::GetSpeed();

    // rot[0] == pitch, rot[1] == heading, rot[2] == bank

    // Change heading based on bank
    rot[1] += sinf(rot[2]) * deltatime;

    // Compute unit length aim vector
    aim[0] = -sinf(rot[1]) * cosf(rot[0]);
    aim[1] = sinf(rot[0]);
    aim[2] = -cosf(rot[1]) * cosf(rot[0]);

    // Apply acceleration, and limit
    speed += accel * deltatime;
    if(speed < -backwardspeedmax)
        speed = -backwardspeedmax;
    if(speed > forwardspeedmax)
        speed = forwardspeedmax;

    // Move the viewer
    pos[0] += aim[0] * speed * deltatime;
    pos[1] += aim[1] * speed * deltatime;
    pos[2] += aim[2] * speed * deltatime;

    // Outer Space avoidance
    if(pos[1] > setheightmax)
        pos[1] = setheightmax;

    // Terrain avoidance
    height = Controller::GetHeight(pos[0], pos[2]);
    height += eyeheightmin;
    if(pos[1] < height)
        pos[1] = height;

    Controller::SetRot(rot);
    Controller::SetPos(pos);
    Controller::SetSpeed(speed);
}


const char *FlyerController::GetName()
{
    return "Flyer";
}


float *FlyerController::GetParamP(int param)
{
    switch(param)
    {
        case pitch_sensitive_Param:
            return &pitchsensitive;
        case bank_sensitive_Param:
            return &banksensitive;
        case pitch_max_Param:
            return &pitchmax;
        case bank_max_Param:
            return &bankmax;
        case accel_sensitive_Param:
            return &accelsensitive;
        case forward_speed_max_Param:
            return &forwardspeedmax;
        case backward_speed_max_Param:
            return &backwardspeedmax;
        case eye_height_min_Param:
            return &eyeheightmin;
        case set_height_max_Param:
            return &setheightmax;
    }

    return 0;
}


void FlyerController::SetParam(int param, float val)
{
    float *paramp;

    paramp = GetParamP(param);
    if(paramp)
        *paramp = val;
    else
        Controller::SetParam(param, val);
}


float FlyerController::GetParam(int param)
{
    float *paramp;

    paramp = GetParamP(param);
    if(paramp)
        return *paramp;
    else
        return Controller::GetParam(param);
}



ExaminerController::ExaminerController(ControlEnvironment& controlenv)
    : Controller(controlenv)
{
    lookat[0] = lookat[1] = lookat[2] = 0;
    inc = 0;
    dist = 1;

    // Initial parameters
    distsensitive = 1;
    pitchsensitive = 1;
    headingsensitive = 1;
}


ExaminerController::~ExaminerController()
{
    ;
}


void ExaminerController::MouseMove(int deltax, int deltay)
{
    float rot[3];

    Controller::GetRot(rot);
    rot[0] -= (float)deltay * pitchsensitive / 100.f;
    rot[1] -= (float)deltax * headingsensitive / 100.f;
    Controller::SetRot(rot);
}


void ExaminerController::MouseButtons(int leftbutton, int middlebutton, int rightbutton)
{
    // Left button causes viewer to move closer
    // Right button to move away
    inc = (float)(rightbutton - leftbutton) * distsensitive;
}


void ExaminerController::MouseWheel(int delta)
{
    // Not used
    ;
}


void ExaminerController::Move(float deltatime)
{
    float rot[3];
    float pos[3];

    Controller::GetRot(rot);

    dist += inc * deltatime;

    // Set position so that it looks at the look-at point
    // given the current camera rotation and distance from
    // the look-at point
    pos[0] = sinf(rot[1]) * cosf(rot[0]) * dist + lookat[0];
    pos[1] = -sinf(rot[0]) * dist + lookat[1];
    pos[2] = cosf(rot[1]) * cosf(rot[0]) * dist + lookat[2];

    Controller::SetPos(pos);
}


void ExaminerController::SetPos(float pos_in[3])
{
    // When manually setting the camera position, compute the
    // matching camera rotation and distance given the current
    // look-at point
    float rot[3];
    float relpos[3] = { pos_in[0] - lookat[0], pos_in[1] - lookat[1], pos_in[2] - lookat[2] };

    dist = (float)sqrt(relpos[0]*relpos[0] + relpos[1]*relpos[1] + relpos[2]*relpos[2]);

    rot[0] = rot[1] = 0;
    if(dist != 0)
        rot[0] = -(float)asin(relpos[1] / dist);
    rot[1] = (float)atan2(relpos[0], relpos[2]);

    Controller::SetPos(pos_in);
    Controller::SetRot(rot);
}


const char *ExaminerController::GetName()
{
    return "Examiner";
}


float *ExaminerController::GetParamP(int param)
{
    switch(param)
    {
        case dist_sensitive_Param:
            return &distsensitive;
        case pitch_sensitive_Param:
            return &pitchsensitive;
        case heading_sensitive_Param:
            return &headingsensitive;
    }

    return 0;
}


void ExaminerController::SetParam(int param, float val)
{
    float *paramp;

    paramp = GetParamP(param);
    if(paramp)
        *paramp = val;
    else
        Controller::SetParam(param, val);
}


float ExaminerController::GetParam(int param)
{
    float *paramp;

    paramp = GetParamP(param);
    if(paramp)
        return *paramp;
    else
        return Controller::GetParam(param);
}


void ExaminerController::SetLookAt(float pos[3])
{
    lookat[0] = pos[0];
    lookat[1] = pos[1];
    lookat[2] = pos[2];
}





ProgController::ProgController(ControlEnvironment& controlenv)
    : Controller(controlenv)
{
    ;
}


ProgController::~ProgController()
{
    ;
}


const char *ProgController::GetName()
{
    return "Programmatic";
}


void ProgController::MouseMove(int deltax, int deltay)
{
    ;
}


void ProgController::MouseButtons(int leftbutton, int middlebutton, int rightbutton)
{
    ;
}


void ProgController::MouseWheel(int delta)
{
    ;
}


void ProgController::Move(float deltatime)
{
    ;
}
