//
// Controller
// An attempt at a unified mouse control system
// jvan006@cs.auckland.ac.nz
//

//
// Basic structure:
//
// There is a control environment on which zero or more controllers can work.
// The control environment keeps the common state.
// The state is modifier by the controller that is current bound to the environment.
// State can be obtained and explicitely set either through the control environment
// or through the current bound controller.
// Controllers modify the environment state by calling the (non-public) SetControl*()
// methods of the environment.
// Calls to the environment SetPos/Rot/Speed() methods will be forwarded to the
// matching methods of the current controller.
// New environment classes must implement the GetHeight() method, which returns the
// height above terrain at a given grid coordinate. This is currently the only
// interaction between controllers and the environment.
// The handling of parameters using enum sucks in the long run, but works.
//

//
// A control interface
// Controllers and control environments implement this interface
//
class ControlInterface
{
    public:
        virtual void MouseMove(int deltax, int deltay) = 0;
        virtual void MouseButtons(int leftbutton, int middlebutton, int rightbutton) = 0;
        virtual void MouseWheel(int delta) = 0;
        virtual void Move(float deltatime) = 0;

        virtual void SetPos(float pos[3]) = 0;
        virtual void SetRot(float rot[3]) = 0;
        virtual void SetSpeed(float speed) = 0;

        virtual void GetPos(float pos[3]) = 0;
        virtual void GetRot(float rot[3]) = 0;
        virtual float GetSpeed() = 0;

        enum { childParam = 128 };
        virtual void SetParam(int param, float val) = 0;
        virtual float GetParam(int param) = 0;
};


//
// A control environment describes the state of any controller
// and the environment in which the controller operates
// It may be shared by multiple controllers
//
class Controller;
class ControlEnvironment : public ControlInterface
{
    public:
        ControlEnvironment();
        virtual ~ControlEnvironment();

        void SetController(Controller& controller);
        Controller& GetController();
        const char *GetControllerName();

        void MouseMoveAbs(int x, int y);
        void CentreMouse(int x, int y);
        void MouseMove(int deltax, int deltay);
        void GetMousePos(int& x, int& y);
        void MouseButtons(int leftbutton, int middlebutton, int rightbutton);
        void MouseWheel(int delta);
        void Move(float deltatime);

        virtual float GetHeight(float x, float z) = 0;

        void SetPos(float pos[3]);
        void SetRot(float rot[3]);
        void SetSpeed(float speed);

        void GetPos(float pos[3]);
        void GetRot(float rot[3]);
        float GetSpeed();

        enum { childParam = ControlInterface::childParam + 256 };
        virtual void SetParam(int param, float val);
        virtual float GetParam(int param);

    protected:
        friend class Controller;

        void SetControlPos(float pos[3]);
        void SetControlRot(float rot[3]);
        void SetControlSpeed(float speed);

    private:
        Controller* curcontrol;
        int mousex, mousey;
        int mousecx, mousecy;

        // State stuff
        float pos[3];            // Viewer position
        float rot[3];            // Viewer rotation
        float speed;             // Viewer speed
};


//
// Generic controller
// A controller gives position, rotation (orientation), and (signed) speed
// It operates in a given control environment
//
class Controller : public ControlInterface
{
    public:
        Controller(ControlEnvironment& controlenv);
        virtual ~Controller();

        virtual void SetPos(float pos[3]);
        virtual void SetRot(float rot[3]);
        virtual void SetSpeed(float speed);

        virtual void GetPos(float pos[3]);
        virtual void GetRot(float rot[3]);
        virtual float GetSpeed();

        virtual const char *GetName() = 0;

        enum { childParam = ControlInterface::childParam + 256 };
        virtual void SetParam(int param, float val);
        virtual float GetParam(int param);

        ControlEnvironment& GetControlEnvironment();

    protected:
        float GetHeight(float x, float z);

    private:
        ControlEnvironment *controlenv;
};


//
// Strider controller
//
class StriderController : public Controller
{
    public:
        StriderController(ControlEnvironment& controlenv);
        virtual ~StriderController();

        void MouseMove(int deltax, int deltay);
        void MouseButtons(int leftbutton, int middlebutton, int rightbutton);
        void MouseWheel(int delta);
        void Move(float deltatime);

        void SetPos(float pos[3]);
        void SetRot(float rot[3]);
        void SetSpeed(float speed);

        const char *GetName();

        void SetParam(int param, float val);
        float GetParam(int param);
        enum { deadzone_Param = Controller::childParam, // Zero speed deadzone size
               speed_sensitive_Param,                   // Speed change sensitivity
               forward_set_speed_max_Param,             // Maximum forward set speed
               backward_set_speed_max_Param,            // Maximum backward set speed (unsigned)
               heading_sensitive_Param,                 // Heading change sensitivity
               height_sensitive_Param,                  // Height change sensitivity
               look_sensitive_Param,                    // Look up/down sensitivity
               forward_speed_max_Param,                 // Maximum forward actual speed
               backward_speed_max_Param,                // Maximum backward actual speed (unsigned)
               set_height_max_Param,                    // Maximum set height
               eye_height_min_Param,                    // Minimum height above terrain
               childParam = Controller::childParam+256 };

    private:
        float *GetParamP(int param);

    private:
        // State stuff
        float reqheight;            // Requested viewer height above datum (not terrain)
        float reqspeed;             // Requested viewer speed (unmodified by scale)
        float inc;                  // Increase (change height)
        int lasty;                  // Last cursor y position with non-zero speed
        float twist;                // Twist of the "head"
        int twisting;               // Changing twist?

        // Parameter stuff
        float deadzone;             // Zero speed deadzone size
        float speedsensitive;       // Speed change sensitivity
        float forwardsetspeedmax;   // Maximum forward set speed
        float backwardsetspeedmax;  // Maximum backward set speed (unsigned)
        float headingsensitive;     // Heading change sensitivity
        float heightsensitive;      // Height change sensitivity
        float looksensitive;        // Look up/down sensitivity
        float forwardspeedmax;      // Maximum forward actual speed
        float backwardspeedmax;     // Maximum backward actual speed (unsigned)
        float setheightmax;         // Maximum set height
        float eyeheightmin;         // Minimum height above terrain
        float twistrestore;         // Twist restore speed
};




//
// Flyer controller
// Behaves like an arcade-style space/aircraft
//
class FlyerController : public Controller
{
    public:
        FlyerController(ControlEnvironment& controlenv);
        virtual ~FlyerController();

        void MouseMove(int deltax, int deltay);
        void MouseButtons(int leftbutton, int middlebutton, int rightbutton);
        void MouseWheel(int delta);
        void Move(float deltatime);

        const char *GetName();

        void SetParam(int param, float val);
        float GetParam(int param);
        enum { pitch_sensitive_Param = Controller::childParam, // Pitch change sensitivity
               bank_sensitive_Param,                           // Bank change sensitivity
               pitch_max_Param,                                // Maximum pitch angle
               bank_max_Param,                                 // Maximum bank angle
               accel_sensitive_Param,                          // Acceleration sensitivity
               forward_speed_max_Param,                        // Maximum forward actual speed
               backward_speed_max_Param,                       // Maximum backward actual speed (unsigned)
               set_height_max_Param,                           // Maximum set height
               eye_height_min_Param,                           // Minimum height above terrain
               childParam = Controller::childParam+256 };

    private:
        float *GetParamP(int param);

    private:
        // State stuff
        float accel;                // Acceleration

        // Parameter stuff
        float pitchsensitive;       // Pitch change sensitivity
        float banksensitive;        // Bank change sensitivity
        float pitchmax;             // Maximum pitch angle
        float bankmax;              // Maximum pitch angle
        float accelsensitive;       // Acceleration sensitivity
        float forwardspeedmax;      // Maximum forward speed
        float backwardspeedmax;     // Maximum backward speed (unsigned)
        float setheightmax;            // Maximum set height
        float eyeheightmin;         // Minimum height above terrain
};


//
// Examiner controller
// Looks at and spins around a point
//
class ExaminerController : public Controller
{
    public:
        ExaminerController(ControlEnvironment& controlenv);
        virtual ~ExaminerController();

        void MouseMove(int deltax, int deltay);
        void MouseButtons(int leftbutton, int middlebutton, int rightbutton);
        void MouseWheel(int delta);
        void Move(float deltatime);

        void SetPos(float pos[3]);

        const char *GetName();

        void SetParam(int param, float val);
        float GetParam(int param);
        enum { dist_sensitive_Param = Controller::childParam, // Distance change sensitivity
               pitch_sensitive_Param,                         // Pitch sensitivity
               heading_sensitive_Param,                       // Heading sensitivity
               childParam = Controller::childParam+256 };

        void SetLookAt(float pos[3]);

    private:
        float *GetParamP(int param);

    private:
        // State stuff
        float lookat[3];            // Look-at point
        float inc;                  // Change in distance
        float dist;                 // Distance of viewer from look-at point

        // Parameter stuff
        float distsensitive;        // Distance change sensitivity
        float pitchsensitive;       // Pitch sensitivity
        float headingsensitive;     // Heading sensitivity

};


//
// Programmatic controller
//
class ProgController : public Controller
{
    public:
        ProgController(ControlEnvironment& controlenv);
        virtual ~ProgController();

        const char *GetName();

        void MouseMove(int deltax, int deltay);
        void MouseButtons(int leftbutton, int middlebutton, int rightbutton);
        void MouseWheel(int delta);
        void Move(float deltatime);

    private:
};
